import siphash
import codecs
import json


class AuthMethod(object):
    SIMPLE_AUTH = 'sa'
    TIMESTAMP_AUTH = 'ta'
    COUNTER_AUTH = 'ca'
    DATA_AUTH = 'da'
    RECURSIVE_DATA_AUTH = 'ra'


class OpenPAYGOMetricsShared(object):

    CONDENSED_KEY_NAMES = {
        # Request
        'serial_number': 'sn',
        'timestamp': 'ts',
        'auth': 'a',
        'request_count': 'rc',
        'data_collection_timestamp': 'dtc',
        'data_format_id': 'df',
        'data_format': 'dfo',
        'data': 'd',
        'historical_data': 'hd',
        'accessories': 'acc',
        # Response
        'token_list': 'tkl',
        'active_until_timestamp': 'auts',
        'active_seconds_left': 'asl',
        'settings': 'st',
        'extra_data': 'ed',
        # Data
        'token_count': 'tc',
        'active_until_timestamp_requested': 'autsr',
        'active_seconds_left_requested': 'aslr'
    }

    @classmethod
    def convert_dict_keys_to_condensed(cls, simple_dict):
        return cls.convert_dict_keys(simple_dict, cls.CONDENSED_KEY_NAMES)
    
    @classmethod
    def convert_dict_keys_to_simple(cls, condensed_dict):
        revert_keys = {v: k for k, v in cls.CONDENSED_KEY_NAMES.items()}
        return cls.convert_dict_keys(condensed_dict, revert_keys)

    @classmethod
    def convert_dict_keys(cls, origin_dict, key_map):
        condensed_dict = {}
        for key in origin_dict:
            if key in key_map:
                condensed_dict[key_map[key]] = origin_dict[key]
            else:
                condensed_dict[key] = origin_dict[key]
        return condensed_dict
    
    @classmethod
    def remove_trailing_empty_elements(cls, list_with_empty):
        while list_with_empty and list_with_empty[-1] is None:
            list_with_empty.pop()
        return list_with_empty

    @classmethod
    def convert_to_metrics_json(cls, data):
        return json.dumps(data, separators=(',', ':'))
    
    @classmethod
    def generate_response_signature_from_data(cls, data, secret_key, serial_number, timestamp=None, request_count=None):
        payload = serial_number
        if timestamp:
            payload += str(timestamp)
        if request_count:
            payload += str(request_count)
        if data.get('active_until_timestamp'):
            payload += str(data.get('active_until_timestamp'))
        if data.get('active_seconds_left'):
            payload += str(data.get('active_seconds_left'))
        if data.get('token_list'):
            payload += cls.convert_to_metrics_json(data.get('token_list'))
        if data.get('settings'):
            payload += cls.convert_to_metrics_json(data.get('settings'))
        if data.get('extra_data'):
            payload += cls.convert_to_metrics_json(data.get('extra_data'))
        return AuthMethod.DATA_AUTH+cls.generate_hash_string(payload, secret_key)
    
    @classmethod
    def generate_request_signature_from_data(cls, data, auth_method, secret_key):
        if auth_method == AuthMethod.SIMPLE_AUTH:
            signature = cls.generate_hash_string(data.get('serial_number'), secret_key)
        elif auth_method == AuthMethod.TIMESTAMP_AUTH:
            if not data.get('timestamp', None):
                raise ValueError('Timestamp is required for Timestamp Auth')
            signature = cls.generate_hash_string(data.get('serial_number') + str(data.get('timestamp')), secret_key)
        elif auth_method == AuthMethod.COUNTER_AUTH:
            if not data.get('request_count', None):
                raise ValueError('Request Count is required for Counter Auth')
            signature = cls.generate_hash_string(data.get('serial_number') + str(data.get('request_count')), secret_key)
        elif auth_method == AuthMethod.DATA_AUTH:
            payload = data.get('serial_number')
            if data.get('timestamp'):
                payload += str(data.get('timestamp'))
            if data.get('request_count'):
                payload += str(data.get('request_count'))
            if data.get('data'):
                payload += cls.convert_to_metrics_json(data.get('data', []))
            if data.get('historical_data'):
                payload += cls.convert_to_metrics_json(data.get('historical_data', []))
            signature = cls.generate_hash_string(payload, secret_key)
        elif auth_method == AuthMethod.RECURSIVE_DATA_AUTH:
            payload = data.get('serial_number')
            payload = cls.generate_hash_string(payload, secret_key)
            if data.get('timestamp'):
                payload = cls.generate_hash_string(payload+str(data.get('timestamp')), secret_key)
            if data.get('request_count'):
                payload = cls.generate_hash_string(payload+str(data.get('request_count')), secret_key)
            payload = cls.generate_hash_string(payload+cls.convert_to_metrics_json(data.get('data', [])), secret_key)
            for time_step_data in data.get('historical_data', []):
                payload = cls.generate_hash_string(payload+cls.convert_to_metrics_json(time_step_data), secret_key)
            signature = payload
        else:
            raise ValueError('Invalid Authentication Method')
        return auth_method+signature


    @classmethod
    def generate_hash_string(cls, input_string, secret_key):
        key = cls.load_secret_key_from_hex(secret_key)
        hash = siphash.SipHash_2_4(key, codecs.encode(input_string, 'utf-8')).hash()
        hash_string = '{:x}'.format(hash)
        return hash_string

    @classmethod
    def load_secret_key_from_hex(cls, secret_key):
        try:
            return codecs.decode(secret_key, 'hex')
        except Exception as e:
            raise ValueError('The secret key provided is not correctly formatted, it should be 32 hexadecimal characters. ')



from .metrics_shared import OpenPAYGOMetricsShared
import copy


class MetricsRequestHandler(object):

    def __init__(self, serial_number, data_format=None, secret_key=None, auth_method=None):
        self.secret_key = secret_key
        self.auth_method = auth_method
        self.request_dict = {
            'serial_number': serial_number,
        }
        self.data_format = data_format
        if self.data_format:
            if self.data_format.get('id'):
                self.request_dict['data_format_id'] = data_format.get('id')
            else:
                self.request_dict['data_format'] = data_format
        self.data = {}
        self.historical_data = {}

    def set_request_count(self, request_count):
        self.request_dict['request_count'] = request_count

    def set_timestamp(self, timestamp):
        self.request_dict['timestamp'] = timestamp

    def set_data(self, data):
        self.data = data

    def set_historical_data(self, historical_data):
        if not self.data_format.get('historical_data_interval'):
            for time_step in historical_data:
                if not time_step.get('timestamp'):
                    raise ValueError('Historical Data objects must have a time stamp if no historical_data_interval is defined.')
        self.historical_data = historical_data

    def get_simple_request_payload(self):
        payload = self.get_simple_request_dict()
        return OpenPAYGOMetricsShared.convert_to_metrics_json(payload)

    def get_simple_request_dict(self):
        simple_request = self.request_dict
        simple_request['data'] = self.data
        simple_request['historical_data'] = self.historical_data
        # We prepare the auth
        if self.auth_method:
            simple_request['auth'] = OpenPAYGOMetricsShared.generate_request_signature_from_data(simple_request, self.auth_method, self.secret_key)
        return simple_request

    def get_condensed_request_payload(self):
        payload = self.get_condensed_request_dict()
        return OpenPAYGOMetricsShared.convert_to_metrics_json(payload)

    def get_condensed_request_dict(self):
        if not self.data_format:
            raise ValueError('No Data Format provided for condensed request')
        data_order = self.data_format.get('data_order')
        if self.data and not data_order:
            raise ValueError('Data Format does not contain data_order')
        historical_data_order = self.data_format.get('historical_data_order')
        if self.historical_data and not historical_data_order:
            raise ValueError('Data Format does not contain historical_data_order')
        condensed_request = copy.deepcopy(self.request_dict)
        condensed_request['data'] = []
        condensed_request['historical_data'] = []
        # We add the data
        data_copy = copy.deepcopy(self.data)
        for var in data_order:
            condensed_request['data'].append(data_copy.pop(var) if var in data_copy else None)
        if len(data_copy) > 0:
            raise ValueError('Additional variables not present in the data format: '+str(data_copy))
        condensed_request['data'] = OpenPAYGOMetricsShared.remove_trailing_empty_elements(condensed_request['data'])
        # We add the historical data
        historical_data_copy = copy.deepcopy(self.historical_data)
        for time_step in historical_data_copy:
            time_step_data = []
            for var in historical_data_order:
                time_step_data.append(time_step.pop(var) if var in time_step else None)
            if len(time_step) > 0:
                raise ValueError('Additional variables not present in the historical data format: '+str(time_step))
            time_step_data = OpenPAYGOMetricsShared.remove_trailing_empty_elements(time_step_data)
            condensed_request['historical_data'].append(time_step_data)
        # We prepare the auth
        if self.auth_method:
            condensed_request['auth'] = OpenPAYGOMetricsShared.generate_request_signature_from_data(condensed_request, self.auth_method, self.secret_key)
        # We replace the key names by the condensed ones
        condensed_request = OpenPAYGOMetricsShared.convert_dict_keys_to_condensed(condensed_request)
        return condensed_request



from .token_shared import OpenPAYGOTokenShared, TokenType
from .token_shared_extended import OpenPAYGOTokenSharedExtended


class OpenPAYGOTokenDecoder(object):
    MAX_TOKEN_JUMP = 64
    MAX_TOKEN_JUMP_COUNTER_SYNC = 100
    MAX_UNUSED_OLDER_TOKENS = 8*2

    @classmethod
    def decode_token(cls, token, secret_key, count, used_counts=None, starting_code=None, value_divider=1, restricted_digit_set=False):
        secret_key = OpenPAYGOTokenShared.load_secret_key_from_hex(secret_key)
        if not starting_code:
            # We generate the starting code from the key if not provided
            starting_code = OpenPAYGOTokenShared.generate_starting_code(secret_key)
        if not restricted_digit_set:
            if len(token) <= 9:
                extended_token = False
            elif len(token) <= 12:
                extended_token = True
            else:
                raise ValueError("Token is too long")
        elif restricted_digit_set:
            if len(token) <= 15:
                extended_token = False
            elif len(token) <= 20:
                extended_token = True
            else:
                raise ValueError("Token is too long")
        token = int(token)
        if not extended_token:
            value, token_type, count, updated_counts = cls.get_activation_value_count_and_type_from_token(token, starting_code, secret_key, count, restricted_digit_set, used_counts)
        else:
            value, token_type, count, updated_counts = cls.get_activation_value_count_from_extended_token(token, starting_code, secret_key, count, restricted_digit_set, used_counts)
        if value and value_divider:
            value = value / value_divider
        return value, token_type, count, updated_counts

    @classmethod
    def get_activation_value_count_and_type_from_token(cls, token, starting_code, key, last_count,
                                                       restricted_digit_set=False, used_counts=None):
        if restricted_digit_set:
            token = OpenPAYGOTokenShared.convert_from_4_digit_token(token)
        valid_older_token = False
        token_base = OpenPAYGOTokenShared.get_token_base(token)  # We get the base of the token
        current_code = OpenPAYGOTokenShared.put_base_in_token(starting_code, token_base)  # We put it into the starting code
        starting_code_base = OpenPAYGOTokenShared.get_token_base(starting_code)  # We get the base of the starting code
        value = cls._decode_base(starting_code_base, token_base)  # If there is a match we get the value from the token
        # We try all combination up until last_count + TOKEN_JUMP, or to the larger jump if syncing counter
        # We could start directly the loop at the last count if we kept the token value for the last count
        if value == OpenPAYGOTokenShared.COUNTER_SYNC_VALUE:
            max_count_try = last_count + cls.MAX_TOKEN_JUMP_COUNTER_SYNC + 1
        else:
            max_count_try = last_count + cls.MAX_TOKEN_JUMP + 1
        for count in range(0, max_count_try):
            masked_token = OpenPAYGOTokenShared.put_base_in_token(current_code, token_base)
            if count % 2:
                if value == OpenPAYGOTokenShared.COUNTER_SYNC_VALUE:
                    this_type = TokenType.COUNTER_SYNC
                elif value == OpenPAYGOTokenShared.PAYG_DISABLE_VALUE:
                    this_type = TokenType.DISABLE_PAYG
                else:
                    this_type = TokenType.SET_TIME
            else:
                this_type = TokenType.ADD_TIME
            if masked_token == token:
                if cls._count_is_valid(count, last_count, value, this_type, used_counts):
                    updated_counts = cls.update_used_counts(used_counts, value, count, this_type)
                    return value, this_type, count, updated_counts
                else:
                    valid_older_token = True
            current_code = OpenPAYGOTokenShared.generate_next_token(current_code, key)  # If not we go to the next token
        if valid_older_token:
            return None, TokenType.ALREADY_USED, None, None
        return None, TokenType.INVALID, None, None

    @classmethod
    def _count_is_valid(cls, count, last_count, value, type, used_counts):
        if value == OpenPAYGOTokenShared.COUNTER_SYNC_VALUE:
            if count > (last_count - cls.MAX_TOKEN_JUMP):
                return True
        elif count > last_count:
            return True
        elif cls.MAX_UNUSED_OLDER_TOKENS > 0:
            if count > last_count - cls.MAX_UNUSED_OLDER_TOKENS:
                if count not in used_counts and type == TokenType.ADD_TIME:
                    return True
        return False

    @classmethod
    def update_used_counts(cls, past_used_counts, value, new_count, type):
        if not past_used_counts:
            return None
        highest_count = max(past_used_counts) if past_used_counts else 0
        if new_count > highest_count:
            highest_count = new_count
        bottom_range = highest_count-cls.MAX_UNUSED_OLDER_TOKENS
        used_counts = []
        if type != TokenType.ADD_TIME or value == OpenPAYGOTokenShared.COUNTER_SYNC_VALUE or value == OpenPAYGOTokenShared.PAYG_DISABLE_VALUE:
            # If it is not an Add-Time token, we mark all the past tokens as used in the range
            for count in range(bottom_range, highest_count+1):
                used_counts.append(count)
        else:
            # If it is an Add-Time token, we just mark the tokens actually used in the range
            for count in range(bottom_range, highest_count+1):
                if count == new_count or count in past_used_counts:
                    used_counts.append(count)
        return used_counts

    @classmethod
    def _decode_base(cls, starting_code_base, token_base):
        decoded_value = token_base - starting_code_base
        if decoded_value < 0:
            return decoded_value + 1000
        else:
            return decoded_value

    @classmethod
    def get_activation_value_count_from_extended_token(cls, token, starting_code, key, last_count,
                                                       restricted_digit_set=False, used_counts=None):
        if restricted_digit_set:
            token = OpenPAYGOTokenSharedExtended.convert_from_4_digit_token(token)
        token_base = OpenPAYGOTokenSharedExtended.get_token_base(token) # We get the base of the token
        current_code = OpenPAYGOTokenSharedExtended.put_base_in_token(starting_code, token_base) # We put it into the starting code
        starting_code_base = OpenPAYGOTokenSharedExtended.get_token_base(starting_code) # We get the base of the starting code
        value = cls._decode_base_extended(starting_code_base, token_base)  # If there is a match we get the value from the token
        max_count_try = last_count + cls.MAX_TOKEN_JUMP + 1
        for count in range(0, max_count_try):
            masked_token = OpenPAYGOTokenSharedExtended.put_base_in_token(current_code, token_base)
            if count % 2:
                this_type = TokenType.SET_TIME
            else:
                this_type = TokenType.ADD_TIME
            if masked_token == token:
                if cls._count_is_valid(count, last_count, value, this_type, used_counts):
                    updated_counts = cls.update_used_counts(used_counts, value, count, this_type)
                    return value, this_type, count, updated_counts
                else:
                    valid_older_token = True
            current_code = OpenPAYGOTokenSharedExtended.generate_next_token(current_code, key) # If not we go to the next token
        if valid_older_token:
            return None, TokenType.ALREADY_USED, None, None
        return None, TokenType.INVALID, None, None

    @classmethod
    def _decode_base_extended(cls, starting_code_base, token_base):
        decoded_value = token_base - starting_code_base
        if decoded_value < 0:
            return decoded_value + 1000000
        else:
            return decoded_value



from .token_encode import OpenPAYGOTokenEncoder
from .token_decode import OpenPAYGOTokenDecoder
from .token_shared import TokenType
from .metrics_request import MetricsRequestHandler
from .metrics_response import MetricsResponseHandler
from .metrics_shared import AuthMethod


def generate_token(**kwargs):
    return OpenPAYGOTokenEncoder.generate_token(**kwargs)


def decode_token(**kwargs):
    return OpenPAYGOTokenDecoder.decode_token(**kwargs)



from .token_shared import OpenPAYGOTokenShared, TokenType
from .token_shared_extended import OpenPAYGOTokenSharedExtended


class OpenPAYGOTokenEncoder(object):

    @classmethod
    def generate_token(cls, secret_key, count, value=None, token_type=TokenType.ADD_TIME, starting_code=None, value_divider=1, restricted_digit_set=False, extended_token=False):
        secret_key = OpenPAYGOTokenShared.load_secret_key_from_hex(secret_key)
        if not starting_code:
            # We generate the starting code from the key if not provided
            starting_code = OpenPAYGOTokenShared.generate_starting_code(secret_key)
        if token_type in [TokenType.ADD_TIME, TokenType.SET_TIME]:
            value = int(round(value * value_divider, 0))
            if not extended_token:
                max_value = OpenPAYGOTokenShared.MAX_ACTIVATION_VALUE
            else:
                max_value = OpenPAYGOTokenSharedExtended.MAX_ACTIVATION_VALUE
            if value > max_value:
                raise ValueError('The value provided is too high.')
        elif value:
            raise ValueError('A value is not allowed for this token type.')
        else:
            if token_type == TokenType.DISABLE_PAYG:
                value = OpenPAYGOTokenShared.PAYG_DISABLE_VALUE
            elif token_type == TokenType.COUNTER_SYNC:
                value = OpenPAYGOTokenShared.COUNTER_SYNC_VALUE
            else:
                raise ValueError('The token type provided is not supported.')
        if extended_token:
            return cls.generate_extended_token(starting_code, secret_key, value, count, token_type, restricted_digit_set)
        else:
            return cls.generate_standard_token(starting_code, secret_key, value, count, token_type, restricted_digit_set)

    @classmethod
    def generate_standard_token(cls, starting_code=None, key=None, value=None, count=None,
                                mode=TokenType.ADD_TIME, restricted_digit_set=False):
        # We get the first 3 digits with encoded value
        starting_code_base = OpenPAYGOTokenShared.get_token_base(starting_code)
        token_base = cls._encode_base(starting_code_base, value)
        current_token = OpenPAYGOTokenShared.put_base_in_token(starting_code, token_base)
        new_count = cls._get_new_count(count, mode)
        for xn in range(0, new_count):
            current_token = OpenPAYGOTokenShared.generate_next_token(current_token, key)
        final_token = OpenPAYGOTokenShared.put_base_in_token(current_token, token_base)
        if restricted_digit_set:
            final_token = OpenPAYGOTokenShared.convert_to_4_digit_token(final_token)
            final_token = '{:015d}'.format(final_token)
        else:
            final_token = '{:09d}'.format(final_token)
        return new_count, final_token

    @classmethod
    def _encode_base(cls, base, number):
        if number + base > 999:
            return number + base - 1000
        else:
            return number + base

    @classmethod
    def generate_extended_token(cls, starting_code, key, value, count, mode=TokenType.ADD_TIME, restricted_digit_set=False):
        starting_code_base = OpenPAYGOTokenSharedExtended.get_token_base(starting_code)
        token_base = cls._encode_base_extended(starting_code_base, value)
        current_token = OpenPAYGOTokenSharedExtended.put_base_in_token(starting_code, token_base)
        new_count = cls._get_new_count(count, mode)
        for xn in range(0, new_count):
            current_token = OpenPAYGOTokenSharedExtended.generate_next_token(current_token, key)
        final_token = OpenPAYGOTokenSharedExtended.put_base_in_token(current_token, token_base)
        if restricted_digit_set:
            final_token = OpenPAYGOTokenSharedExtended.convert_to_4_digit_token(final_token)
            final_token = '{:020d}'.format(final_token)
        else:
            final_token = '{:012d}'.format(final_token)
        return new_count, final_token

    @classmethod
    def _encode_base_extended(cls, base, number):
        if number + base > 999999:
            return number + base - 1000000
        else:
            return number + base
        
    @classmethod
    def _get_new_count(cls, count, mode):
        current_count_odd = count % 2
        if mode in [TokenType.SET_TIME, TokenType.DISABLE_PAYG, TokenType.COUNTER_SYNC]:
            if current_count_odd: # Odd numbers are for Set Time, Disable PAYG or Counter Sync
                new_count = count+2
            else:
                new_count = count+1
        else:
            if current_count_odd: # Even numbers are for Add Time
                new_count = count+1
            else:
                new_count = count+2
        return new_count



import json
from .metrics_shared import OpenPAYGOMetricsShared
import copy
from datetime import datetime, timedelta

class MetricsResponseHandler(object):

    def __init__(self, received_metrics, data_format=None, secret_key=None, last_request_count=None, last_request_timestamp=None):
        self.received_metrics = received_metrics
        self.request_dict = json.loads(received_metrics)
        # We convert the base variable names to simple
        self.request_dict = OpenPAYGOMetricsShared.convert_dict_keys_to_simple(self.request_dict)
        # We add the reception timestamp if not timestamp was provided
        self.request_timestamp = self.request_dict.get('timestamp')
        if not self.request_dict.get('timestamp'):
            self.timestamp = int(datetime.now().timestamp())
        else:
            self.timestamp = self.request_dict.get('timestamp')
        self.response_dict = {}
        self.secret_key = secret_key
        self.data_format = data_format
        self.last_request_count = last_request_count
        self.last_request_timestamp = last_request_timestamp
        if not self.data_format and self.request_dict.get('data_format'):
            self.data_format = self.request_dict.get('data_format')

    def get_device_serial(self):
        return self.request_dict.get('serial_number')
    
    def get_data_format_id(self):
        return self.request_dict.get('data_format_id')
    
    def data_format_available(self):
        return self.data_format != None
    
    def set_device_parameters(self, secret_key=None, data_format=None, last_request_count=None, last_request_timestamp=None):
        if secret_key:
            self.secret_key = secret_key
        if data_format:
            self.data_format = data_format
        if last_request_count:
            self.last_request_count = last_request_count
        if last_request_timestamp:
            self.last_request_timestamp = last_request_timestamp

    def is_auth_valid(self):
        auth_string = self.request_dict.get('auth', None)
        if not auth_string:
            return True
        elif not self.secret_key:
            raise ValueError('Secret key is required to check the auth.')
        self.auth_method = auth_string[:2]
        new_signature = OpenPAYGOMetricsShared.generate_request_signature_from_data(self.request_dict, self.auth_method, self.secret_key)
        if auth_string == new_signature:
            request_count = self.request_dict.get('request_count')
            if request_count and self.last_request_count and request_count <= self.last_request_count:
                return False
            timestamp = self.request_dict.get('timestamp')
            if timestamp and self.last_request_timestamp and timestamp <= self.last_request_timestamp:
                return False
            return True
        return False

    def get_simple_metrics(self):
        # We start the process by making a copy of the dict to work with
        simple_dict = copy.deepcopy(self.request_dict)
        simple_dict.pop('auth') if 'auth' in simple_dict else None # We remove the auth
        # We process the data and replace it
        simple_dict['data'] = self._get_simple_data()
        # We process the historical data
        simple_dict['historical_data'] = self._get_simple_historical_data()
        # We fill in the timestamps for each time step
        simple_dict['historical_data'] = self._fill_timestamp_in_historical_data(simple_dict['historical_data'])
        return simple_dict
    
    def get_data_timestamp(self):
        return self.request_dict.get('data_collection_timestamp', self.timestamp)
    
    def get_request_timestmap(self):
        return self.request_timestamp
    
    def get_request_count(self):
        return self.request_dict.get('request_count')
    
    def get_token_count(self):
        data = self._get_simple_data()
        return data.get('token_count')
    
    def expects_token_answer(self):
        return self.get_token_count() is not None

    def add_tokens_to_answer(self, token_list):
        self.response_dict['token_list'] = token_list

    def expects_time_answer(self):
        data = self._get_simple_data()
        if data.get('active_until_timestamp_requested', False) or data.get('active_seconds_left_requested', False):
            return True
        return False

    def add_time_to_answer(self, target_datetime):
        data = self._get_simple_data()
        if data.get('active_until_timestamp_requested', False):
            self.response_dict['active_until_timestamp'] = target_datetime.timestamp()
        elif data.get('active_seconds_left_requested', False):
            self.response_dict['active_seconds_left'] = (datetime.now() - target_datetime).total_seconds()
        else:
            raise ValueError('No time requested')
        
    def add_new_base_url_to_answer(self, new_base_url):
        self.add_settings_to_answer({'base_url': new_base_url})

    def add_settings_to_answer(self, settings_dict):
        if not self.response_dict.get('settings'):
            self.response_dict['settings'] = {}
        self.response_dict['settings'].update(settings_dict)

    def add_extra_data_to_answer(self, extra_data_dict):
        if not self.response_dict.get('extra_data'):
            self.response_dict['extra_data'] = {}
        self.response_dict['extra_data'].update(extra_data_dict)

    def get_answer_payload(self):
        payload = self.get_answer_dict()
        return OpenPAYGOMetricsShared.convert_to_metrics_json(payload)

    def get_answer_dict(self):
        # If there is not data format, we just return the full response
        condensed_answer = copy.deepcopy(self.response_dict)
        condensed_answer['auth'] = OpenPAYGOMetricsShared.generate_response_signature_from_data(
            serial_number=self.request_dict.get('serial_number'),
            request_count=self.request_dict.get('request_count'),
            data=condensed_answer,
            timestamp=self.request_dict.get('timestamp'),
            secret_key=self.secret_key
        )
        return OpenPAYGOMetricsShared.convert_dict_keys_to_condensed(condensed_answer)

    def _get_simple_data(self):
        data = copy.deepcopy(self.request_dict.get('data'))
        # If no data or not condensed in list, we just return it
        if not data:
            return {}
        if not isinstance(data, list):
            return data
        data_order = self.data_format.get('data_order')
        if not data_order:
            raise ValueError('Data Format does not contain data_order')
        clean_data = {}
        data_len = len(data)
        for idx, var in enumerate(data_order):
            clean_data[var] = data[idx] if idx < data_len else None
        data = data[data_len:]
        if len(data) > 0:
            raise ValueError('Additional variables not present in the data format: '+str(data))
        return OpenPAYGOMetricsShared.convert_dict_keys_to_simple(clean_data)
    
    def _get_simple_historical_data(self):
        historical_data = copy.deepcopy(self.request_dict.get('historical_data'))
        if not historical_data:
            return []
        historical_data_order = self.data_format.get('historical_data_order')
        clean_historical_data = []
        for time_step in historical_data:
            time_step_data = {}
            if isinstance(time_step, list):
                if not historical_data_order:
                    raise ValueError('Data Format does not contain historical_data_order')
                timse_step_len = len(time_step)
                for idx, var in enumerate(historical_data_order):
                    if idx < timse_step_len:
                        time_step_data[var] = time_step[idx]
                time_step = time_step[timse_step_len:]
                if len(time_step) > 0:
                    raise ValueError('Additional variables not present in the historical data format: '+str(time_step))
            elif isinstance(time_step, dict):
                for key in time_step:
                    if key.isdigit() and int(key) < len(historical_data_order):
                        time_step_data[historical_data_order[int(key)]] = time_step[key]
                    else:
                        time_step_data[key] = time_step[key]
            else:
                raise ValueError('Invalid historical data step type: '+str(time_step))
            clean_historical_data.append(time_step_data)
        return clean_historical_data
    
    def _fill_timestamp_in_historical_data(self, historical_data):
        last_timestamp = datetime.fromtimestamp(self.get_data_timestamp())
        for idx, time_step in enumerate(historical_data):
            if time_step.get('relative_time') is not None:
                last_timestamp = last_timestamp + timedelta(seconds=int(time_step.get('relative_time')))
                historical_data[idx]['timestamp'] = int(last_timestamp.timestamp())
                del historical_data[idx]['relative_time']
            elif time_step.get('timestamp'):
                last_timestamp = datetime.fromtimestamp(time_step.get('timestamp'))
            else:
                if idx != 0:
                    last_timestamp = last_timestamp + timedelta(seconds=int(self.data_format.get('historical_data_interval')))
                historical_data[idx]['timestamp'] = int(last_timestamp.timestamp())
        return historical_data



from setuptools import setup, find_packages


setup(
    name="openpaygo", 
    packages=find_packages(),
    version='0.5.0',
    license='MIT',
    author="Solaris Offgrid",
    url='https://github.com/EnAccess/OpenPAYGO-python/',
    install_requires=[
        'siphash',
    ],
)